// Feature check for 'PQXX_HAVE_RANGES'.
// Generated by generate_cxx_checks.py.
#include <version>
#if !defined(__cpp_lib_ranges)
#  error "No PQXX_HAVE_RANGES: __cpp_lib_ranges is not set."
#endif
#if !__cpp_lib_ranges
#  error "No PQXX_HAVE_RANGES: __cpp_lib_ranges is false."
#endif

int main() {}
