// Feature check for 'PQXX_HAVE_MULTIDIM'.
// Generated by generate_cxx_checks.py.
#include <version>
#if !defined(__cpp_multidimensional_subscript)
#  error "No PQXX_HAVE_MULTIDIM: __cpp_multidimensional_subscript is not set."
#endif
#if !__cpp_multidimensional_subscript
#  error "No PQXX_HAVE_MULTIDIM: __cpp_multidimensional_subscript is false."
#endif

#include <source_location>


struct foo
{
  int operator[](int i) const { return i; }
  int operator[](int i, std::source_location) const { return i - 1; }
};


int main()
{
  return foo{}[1, std::source_location::current()];
}
